use std::{
  ops::{Deref, DerefMut},
  time::Duration,
};

use color_eyre::eyre::Result;
use crossterm::{
  cursor,
  event::{
    DisableBracketedPaste, DisableMouseCapture, EnableBracketedPaste, EnableMouseCapture,
    Event as CrosstermEvent, KeyEvent, KeyEventKind, MouseEvent,
  },
  terminal::SetTitle,
  terminal::{EnterAlternateScreen, LeaveAlternateScreen},
};
use futures::{FutureExt, StreamExt};
use ratatui::backend::CrosstermBackend as Backend;
use serde::{Deserialize, Serialize};
use tokio::{
  sync::mpsc::{self, UnboundedReceiver, UnboundedSender},
  task::JoinHandle,
};
use tokio_util::sync::CancellationToken;

pub type IO = std::io::Stdout;
pub fn io() -> IO {
  std::io::stdout()
}
pub type Frame<'a> = ratatui::Frame<'a>;

fn clear_terminal_for_full_redraw<B: ratatui::backend::Backend>(
  terminal: &mut ratatui::Terminal<B>,
) -> std::result::Result<(), B::Error> {
  // Terminal::clear() queries the cursor position first, but detached PTYs (such as Docker's
  // `-t`) have no terminal emulator to answer that query.
  terminal.backend_mut().clear_region(ratatui::backend::ClearType::All)?;
  terminal.current_buffer_mut().reset();
  terminal.swap_buffers();
  Ok(())
}

#[derive(Clone, Debug, Serialize, Deserialize)]
pub enum Event {
  Init,
  Quit,
  Error,
  Closed,
  Tick,
  Render,
  FocusGained,
  FocusLost,
  Paste(String),
  Key(KeyEvent),
  Mouse(MouseEvent),
  Resize(u16, u16),
}

pub struct Tui {
  pub terminal: ratatui::Terminal<Backend<IO>>,
  pub task: JoinHandle<()>,
  pub cancellation_token: CancellationToken,
  pub event_rx: UnboundedReceiver<Event>,
  pub event_tx: UnboundedSender<Event>,
  pub frame_rate: f64,
  pub tick_rate: f64,
  pub mouse: bool,
  pub paste: bool,
  pub title: Option<String>,
}

impl Tui {
  pub fn new() -> Result<Self> {
    let tick_rate = 4.0;
    let frame_rate = 15.0;
    let terminal = ratatui::Terminal::new(Backend::new(io()))?;
    let (event_tx, event_rx) = mpsc::unbounded_channel();
    let cancellation_token = CancellationToken::new();
    let task = tokio::spawn(async {});
    let mouse = true;
    let paste = true;
    Ok(Self {
      terminal,
      task,
      cancellation_token,
      event_rx,
      event_tx,
      frame_rate,
      tick_rate,
      mouse,
      paste,
      title: None,
    })
  }

  pub fn tick_rate(mut self, tick_rate: Option<f64>) -> Self {
    if let Some(tick_rate) = tick_rate {
      self.tick_rate = tick_rate;
    };
    self
  }

  pub fn frame_rate(mut self, frame_rate: Option<f64>) -> Self {
    if let Some(frame_rate) = frame_rate {
      self.frame_rate = frame_rate;
    }
    self
  }

  pub fn mouse(mut self, mouse: Option<bool>) -> Self {
    if let Some(mouse) = mouse {
      self.mouse = mouse;
    }
    self
  }

  pub fn paste(mut self, paste: bool) -> Self {
    self.paste = paste;
    self
  }

  pub fn title(mut self, title: Option<String>) -> Self {
    self.title = title;
    self
  }

  pub fn start(&mut self) {
    let tick_delay = std::time::Duration::from_secs_f64(1.0 / self.tick_rate);
    let render_delay = std::time::Duration::from_secs_f64(1.0 / self.frame_rate);
    self.cancel();
    self.cancellation_token = CancellationToken::new();
    let _cancellation_token = self.cancellation_token.clone();
    let _event_tx = self.event_tx.clone();
    self.task = tokio::spawn(async move {
      let mut reader = crossterm::event::EventStream::new();
      let mut tick_interval = tokio::time::interval(tick_delay);
      let mut render_interval = tokio::time::interval(render_delay);
      let mut last_frame_mouse_event: Option<MouseEvent> = None; // debounce mouse events
      _event_tx.send(Event::Init).unwrap();
      loop {
        let tick_delay = tick_interval.tick();
        let render_delay = render_interval.tick();
        let crossterm_event = reader.next().fuse();
        tokio::select! {
          _ = _cancellation_token.cancelled() => {
            break;
          }
          maybe_event = crossterm_event => {
            match maybe_event {
              Some(Ok(evt)) => {
                match evt {
                  CrosstermEvent::Key(key) => {
                    if key.kind == KeyEventKind::Press {
                      _event_tx.send(Event::Key(key)).unwrap();
                    }
                  },
                  CrosstermEvent::Mouse(mouse) => {
                     last_frame_mouse_event = Some(mouse);
                  },
                  CrosstermEvent::Resize(x, y) => {
                    _event_tx.send(Event::Resize(x, y)).unwrap();
                  },
                  CrosstermEvent::FocusLost => {
                    _event_tx.send(Event::FocusLost).unwrap();
                  },
                  CrosstermEvent::FocusGained => {
                    _event_tx.send(Event::FocusGained).unwrap();
                  },
                  CrosstermEvent::Paste(s) => {
                    _event_tx.send(Event::Paste(s)).unwrap();
                  },
                }
              }
              Some(Err(_)) => {
                _event_tx.send(Event::Error).unwrap();
              }
              None => {},
            }
          },
          _ = tick_delay => {
            _event_tx.send(Event::Tick).unwrap();
          },
          _ = render_delay => {
            _event_tx.send(Event::Render).unwrap();
            if let Some(mouse_event) = last_frame_mouse_event.take() {
              _event_tx.send(Event::Mouse(mouse_event)).unwrap();
            }
          },
        }
      }
    });
  }

  pub fn stop(&self) -> Result<()> {
    self.cancel();
    let mut counter = 0;
    while !self.task.is_finished() {
      std::thread::sleep(Duration::from_millis(1));
      counter += 1;
      if counter > 50 {
        self.task.abort();
      }
      if counter > 100 {
        log::error!("Failed to abort task in 100 milliseconds for unknown reason");
        break;
      }
    }
    Ok(())
  }

  pub fn enter(&mut self) -> Result<()> {
    crossterm::terminal::enable_raw_mode()?;
    if let Some(title) = &self.title {
      crossterm::execute!(io(), SetTitle(title))?;
    }
    crossterm::execute!(io(), EnterAlternateScreen, cursor::Hide)?;
    clear_terminal_for_full_redraw(&mut self.terminal)?;
    if self.mouse {
      crossterm::execute!(io(), EnableMouseCapture)?;
    }
    if self.paste {
      crossterm::execute!(io(), EnableBracketedPaste)?;
    }
    self.start();
    Ok(())
  }

  pub fn exit(&mut self) -> Result<()> {
    self.stop()?;
    if crossterm::terminal::is_raw_mode_enabled()? {
      self.flush()?;
      if self.paste {
        crossterm::execute!(io(), DisableBracketedPaste)?;
      }
      if self.mouse {
        crossterm::execute!(io(), DisableMouseCapture)?;
      }
      crossterm::execute!(io(), LeaveAlternateScreen, cursor::Show)?;
      crossterm::terminal::disable_raw_mode()?;
    }
    Ok(())
  }

  pub fn cancel(&self) {
    self.cancellation_token.cancel();
  }

  pub fn resume(&mut self) -> Result<()> {
    self.enter()?;
    Ok(())
  }

  pub async fn next(&mut self) -> Option<Event> {
    self.event_rx.recv().await
  }
}

impl Deref for Tui {
  type Target = ratatui::Terminal<Backend<IO>>;

  fn deref(&self) -> &Self::Target {
    &self.terminal
  }
}

impl DerefMut for Tui {
  fn deref_mut(&mut self) -> &mut Self::Target {
    &mut self.terminal
  }
}

impl Drop for Tui {
  fn drop(&mut self) {
    self.exit().unwrap();
  }
}

#[cfg(test)]
mod tests {
  use super::*;
  use ratatui::backend::{Backend as _, ClearType, TestBackend};

  #[test]
  fn resumed_terminal_redraws_an_unchanged_frame() {
    let mut terminal = ratatui::Terminal::new(TestBackend::new(8, 1)).unwrap();
    terminal.draw(|frame| frame.render_widget("rainfrog", frame.area())).unwrap();
    terminal.backend().assert_buffer_lines(["rainfrog"]);

    // Entering a fresh alternate screen clears the physical display without
    // changing Ratatui's cached previous frame.
    terminal.backend_mut().clear_region(ClearType::All).unwrap();
    terminal.backend().assert_buffer_lines(["        "]);

    clear_terminal_for_full_redraw(&mut terminal).unwrap();
    terminal.draw(|frame| frame.render_widget("rainfrog", frame.area())).unwrap();

    terminal.backend().assert_buffer_lines(["rainfrog"]);
  }
}
